
# This file was generated by 'versioneer.py' (0.28) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2026-06-30T17:28:47-0500",
 "dirty": true,
 "error": null,
 "full-revisionid": "aacf44424ddb90580c338de35f04be65c07928ea",
 "version": "0.66.0+0.gaacf44424.dirty"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
